#
# SPDX-FileCopyrightText: Copyright (c) 2021-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Buffer to store paths during their computation"""

import mitsuba as mi
import drjit as dr
from typing import Union, Type

from sionna.rt.constants import InteractionType, INVALID_SHAPE,\
    INVALID_PRIMITIVE
from sionna.rt.utils import theta_phi_from_unit_vec


class PathsBuffer:
    r"""
    Class used to store the paths during their computation

    The output of a path solver is an instance of this class from which an
    instance of :class:`~sionna.rt.Paths` is built.

    :param buffer_size: Size of the buffer
    :param max_depth: Maximum depth
    """

    def __init__(self, buffer_size : int, max_depth : int):

        # Size of the array
        # If `max_depth` is 0, we allocate one element to store paths data
        depth_dim_size = max(1, max_depth)

        self._max_depth = max_depth
        self._buffer_size = buffer_size
        self._depth_dim_size = depth_dim_size

        # Effective number of paths. This counter should be used to count the
        # number of paths effectively found by a solver.
        # Note that the buffer can be shrinked to this value using self.shrink()
        self._paths_counter = mi.UInt(0)

        # Set to True for if the path is valid
        self._valid = dr.full(dr.mask_t(mi.Float), False, buffer_size)

        # Index of the source from which the path originates
        self._src_indices = dr.zeros(mi.UInt, buffer_size)

        # Index of the target to which the path connects to
        self._tgt_indices = dr.zeros(mi.UInt, buffer_size)

        # Angles of arrival and departure
        self._theta_t = dr.zeros(mi.Float, buffer_size)
        self._phi_t = dr.zeros(mi.Float, buffer_size)
        #
        self._theta_r = dr.zeros(mi.Float, buffer_size)
        self._phi_r = dr.zeros(mi.Float, buffer_size)

        # Type of interaction (specular reflection, diffuse reflection, etc)
        self._interaction_types = dr.full(mi.TensorXu, InteractionType.NONE,
                                          [buffer_size, depth_dim_size])

        # Path vertices
        # The additional two entries along the depth dimension correspond to the
        # endpoints of the paths.
        self._vertices_x = dr.zeros(mi.TensorXf, [buffer_size,depth_dim_size])
        self._vertices_y = dr.zeros(mi.TensorXf, [buffer_size,depth_dim_size])
        self._vertices_z = dr.zeros(mi.TensorXf, [buffer_size,depth_dim_size])

        # Pointers to intersected shapes
        # Shapes pointers are stored as unsigned integers
        self._shapes = dr.full(mi.TensorXu, INVALID_SHAPE,
                                   [buffer_size, depth_dim_size])

        # Index of intersected primitive of the shape
        self._primitives = dr.full(mi.TensorXu, INVALID_PRIMITIVE,
                                   [buffer_size, depth_dim_size])

        # Channel inpulse response coefficients and delays are initialized to
        # `None`
        self._a = None
        self._tau = None
        # Doppler shifts of paths are initialized to `None`
        self._doppler = None

    @property
    def max_depth(self):
        r"""Maximum depth

        :type: :py:class:`int`
        """
        return self._max_depth

    @property
    def depth_dim_size(self):
        r"""Size of the depth dimension of the arrays and tensors instantiated
        by this class. Equals to 1 if ``max_depth == 0`` or to ``max_depth``
        otherwise.

        :type: :py:class:`int`
        """
        return self._depth_dim_size

    @property
    def buffer_size(self):
        r"""Size of the buffer

        :type: :py:class:`int`
        """
        return self._buffer_size

    @property
    def paths_counter(self):
        r"""Number of paths stored in the buffer

        :type: :py:class:`int`
        """
        return self._paths_counter

    @property
    def valid(self):
        r"""Flag indicating valid paths

        :type: :py:class:`mi.Bool`
        """
        return self._valid

    @valid.setter
    def valid(self, v):
        self._valid = v

    @property
    def source_indices(self):
        r"""Indices of the source from which the paths originates

        :type: :py:class:`mi.UInt`
        """
        return self._src_indices

    @property
    def target_indices(self):
        r"""Indices of the target to which the paths connect

        :type: :py:class:`mi.UInt`
        """
        return self._tgt_indices

    @property
    def theta_t(self):
        r"""Zenith  angles of departure [rad]

        :type: :py:class:`mi.Float`
        """
        return self._theta_t

    @property
    def phi_t(self):
        r"""Azimuth  angles of departure [rad]

        :type: :py:class:`mi.Float`
        """
        return self._phi_t

    @property
    def theta_r(self):
        r"""Zenith  angles of arrival [rad]

        :type: :py:class:`mi.Float`
        """
        return self._theta_r

    @property
    def phi_r(self):
        r"""Azimuth angles of arrival [rad]

        :type: :py:class:`mi.Float`
        """
        return self._phi_r

    @property
    def interaction_types(self):
        r"""Interactions types represented using
        :data:`~sionna.rt.constants.InteractionType`

        :type: :py:class:`mi.TensorXu [buffer_size, depth_dim_size]`
        """
        return self._interaction_types

    @property
    def vertices_x(self):
        r"""X coordinates of the paths' vertices

        :type: :py:class:`mi.TensorXf [buffer_size, depth_dim_size]`
        """
        return self._vertices_x

    @property
    def vertices_y(self):
        r"""Y coordinates of the paths' vertices

        :type: :py:class:`mi.TensorXf [buffer_size, depth_dim_size]`
        """
        return self._vertices_y

    @property
    def vertices_z(self):
        r"""Z coordinates of the paths' vertices

        :type: :py:class:`mi.TensorXf [buffer_size, depth_dim_size]`
        """
        return self._vertices_z

    @property
    def shapes(self):
        r"""Intersected shapes. Invalid shapes are represented by
        :data:`~sionna.rt.constants.INVALID_SHAPE`

        :type: :py:class:`mi.TensorXu [buffer_size, depth_dim_size]`
        """
        return self._shapes

    @property
    def primitives(self):
        r"""Intersected primitives. Invalid primitives are represented by
        :data:`~sionna.rt.constants.INVALID_PRIMITIVE`.

        :type: :py:class:`mi.TensorXu [buffer_size, depth_dim_size]`
        """
        return self._primitives

    @property
    def a(self):
        r"""Paths coefficients for every receive antenna pattern and transmit
        antenna pattern. ``a[n][m]`` stores the array of paths coefficients for
        the ``n`` th receive antenna pattern and the ``m`` th transmit antenna
        pattern

        :type: :py:class:`Tuple[Tuple[mi.Complex2f]]`
        """
        return self._a

    @a.setter
    def a(self, v):
        self._a = v

    @property
    def tau(self):
        r"""Paths delays [s]

        :type: :py:class:`mi.Float`:
        """
        return self._tau

    @tau.setter
    def tau(self, v):
        self._tau = v

    @property
    def doppler(self):
        r""" Paths Doppler shifts [Hz]

        :type: :py:class:`mi.Float`
        """
        return self._doppler

    @doppler.setter
    def doppler(self, v):
        self._doppler = v


    def schedule(self) -> None:
        # TODO: let DrJit walk through this class on its own, e.g. w/ @dataclass
        dr.schedule(
            self._max_depth,
            self._buffer_size,
            self._depth_dim_size,
            self._paths_counter,
            self._valid,
            self._src_indices,
            self._tgt_indices,
            self._theta_t,
            self._phi_t,
            self._theta_r,
            self._phi_r,
            self._interaction_types,
            self._vertices_x,
            self._vertices_y,
            self._vertices_z,
            self._shapes,
            self._primitives,
            self._a,
            self._tau,
            self._doppler,
        )

    @dr.syntax
    def add_paths(self,
                  depth : mi.UInt,
                  indices : mi.UInt,
                  sample_data : int,
                  valid : mi.Bool,
                  tgt_index : mi.UInt,
                  k_tx : mi.Vector3f,
                  k_rx : mi.Vector3f,
                  active : mi.Bool) -> None:
        # pylint: disable=line-too-long
        r"""
        Adds paths to the buffer

        :param depth: Depths of the paths to add
        :param indices: Indices where to add paths in the buffer
        :param sample_data: Paths data to store
        :param valid: Flags indicating if the paths are valid, i.e., if their compute are finalized
        :param tgt_index: Targets to which the paths connect
        :param k_tx: Directions of departure of the paths
        :param k_rx: Directions of arrival of the paths
        :param active: Flags specifying active paths. Inactive paths are not added.
        """

        depth_dim_size = self._depth_dim_size

        # Update the valid flag
        dr.scatter(self._valid, valid, indices, active=active)
        # Update the source index
        dr.scatter(self._src_indices, sample_data.src_indices, indices, active)
        # Update the target index
        dr.scatter(self._tgt_indices, tgt_index, indices, active)
        # Update direction of departure
        theta_t, phi_t = theta_phi_from_unit_vec(k_tx)
        dr.scatter(self._theta_t, theta_t, indices, active)
        dr.scatter(self._phi_t, phi_t, indices, active)
        # Update direction of arrival
        theta_r, phi_r = theta_phi_from_unit_vec(k_rx)
        dr.scatter(self._theta_r, theta_r, indices, active)
        dr.scatter(self._phi_r, phi_r, indices, active)

        # Only add additional path data if `max_depth > 0`
        d = dr.ones(mi.UInt, dr.width(active))
        while d <= depth:
            interaction_types, shapes, primitives, vertices\
                = sample_data.get(d)

            # Indices for updating the interation type, shape, and primitive
            # arrays
            indices_t = indices*depth_dim_size + d - 1

            # Update interaction types
            dr.scatter(self._interaction_types.array, interaction_types,
                        indices_t, active)
            # Update shapes
            dr.scatter(self._shapes.array, shapes, indices_t, active)
            # Update primitives
            dr.scatter(self._primitives.array, primitives, indices_t,active)
            # Update interaction types
            dr.scatter(self._vertices_x.array, vertices.x, indices_t,active)
            dr.scatter(self._vertices_y.array, vertices.y, indices_t,active)
            dr.scatter(self._vertices_z.array, vertices.z, indices_t,active)

            d += 1

    def shrink(self) -> None:
        r"""
        Shrinks the buffer size to :attr:`~sionna.rt.PathsBuffer.path_counter`

        Only the first :attr:`~sionna.rt.PathsBuffer.path_counter` items are
        kept.
        """

        # Effective number of paths
        num_paths = dr.minimum(dr.max(self._paths_counter),
                               self._buffer_size)[0]

        depth_dim_size = self._depth_dim_size

        self._valid = dr.reshape(mi.Bool, self._valid, num_paths,
                                 shrink=True)
        self._src_indices = dr.reshape(mi.UInt, self._src_indices,
                                       num_paths, shrink=True)
        self._tgt_indices = dr.reshape(mi.UInt, self._tgt_indices,
                                       num_paths, shrink=True)
        if self._a is not None:
            a = []
            for a_ in self._a:
                a1 = []
                for a__ in a_:
                    a1.append(dr.reshape(mi.Complex2f, a__, num_paths,
                                         shrink=True))
                a.append(a1)
            self._a = a
        if self._tau is not None:
            self._tau = dr.reshape(mi.Float, self._tau, num_paths, shrink=True)
        if self._doppler is not None:
            self._doppler = dr.reshape(mi.Float, self._doppler, num_paths,
                                       shrink=True)
        self._theta_t = dr.reshape(mi.Float, self._theta_t,
                                   num_paths, shrink=True)
        self._phi_t = dr.reshape(mi.Float, self._phi_t,
                                 num_paths, shrink=True)
        self._theta_r = dr.reshape(mi.Float, self._theta_r,
                                   num_paths, shrink=True)
        self._phi_r = dr.reshape(mi.Float, self._phi_r,
                                 num_paths, shrink=True)
        self._interaction_types = mi.TensorXu(
            dr.reshape(mi.UInt, self._interaction_types.array,
                       num_paths*depth_dim_size, shrink=True),
            shape=(num_paths,depth_dim_size)
        )
        self._vertices_x = mi.TensorXf(
            dr.reshape(mi.Float, self._vertices_x.array,
                       num_paths*depth_dim_size, shrink=True),
            shape=(num_paths, depth_dim_size)
        )
        self._vertices_y = mi.TensorXf(
            dr.reshape(mi.Float, self._vertices_y.array,
                       num_paths*depth_dim_size, shrink=True),
            shape=(num_paths, depth_dim_size)
        )
        self._vertices_z = mi.TensorXf(
            dr.reshape(mi.Float, self._vertices_z.array,
                       num_paths*depth_dim_size, shrink=True),
            shape=(num_paths, depth_dim_size)
        )
        self._shapes = mi.TensorXu(
            dr.reshape(mi.UInt, self._shapes.array,
                       num_paths*depth_dim_size, shrink=True),
            shape=(num_paths, depth_dim_size)
        )
        self._primitives = mi.TensorXu(
            dr.reshape(mi.UInt, self._primitives.array,
                       num_paths*depth_dim_size, shrink=True),
            shape=(num_paths, depth_dim_size)
        )

        self._buffer_size = num_paths

    def discard_invalid(self) -> None:
        r"""
        Discards invalid paths

        This function discards paths for which the entry in
        :attr:`~sionna.rt.PathsBuffer.valid` is set to `False`.
        """

        depth_dim_size = self._depth_dim_size
        self.schedule()

        # Indices of valid paths
        valid_ind = dr.compress(self._valid)

        # Number of valid paths
        num_valid_paths = dr.shape(valid_ind)[0]

        # Tensor indices for gathering the paths data
        depth_ind = dr.tile(dr.arange(mi.UInt, 0, depth_dim_size),
                            num_valid_paths)
        # Need to handle the case where no paths is found separately
        if num_valid_paths == 0:
            tensor_gind = []
        else:
            tensor_gind = dr.repeat(valid_ind, depth_dim_size)*depth_dim_size\
                            + depth_ind

        self._valid = dr.gather(mi.Bool, self._valid, valid_ind)
        self._src_indices = dr.gather(mi.UInt, self._src_indices, valid_ind)
        self._tgt_indices = dr.gather(mi.UInt, self._tgt_indices, valid_ind)
        if self._a is not None:
            a = []
            for a_ in self._a:
                a1 = []
                for a__ in a_:
                    a1.append(dr.gather(mi.Complex2f, a__, valid_ind))
                a.append(a1)
            self._a = a
        if self._tau is not None:
            self._tau = dr.gather(mi.Float, self._tau, valid_ind)
        if self._doppler is not None:
            self._doppler = dr.gather(mi.Float, self._doppler, valid_ind)
        self._theta_t = dr.gather(mi.Float, self._theta_t, valid_ind)
        self._phi_t = dr.gather(mi.Float, self._phi_t, valid_ind)
        self._theta_r = dr.gather(mi.Float, self._theta_r, valid_ind)
        self._phi_r = dr.gather(mi.Float, self._phi_r, valid_ind)
        self._interaction_types = mi.TensorXu(
            dr.gather(mi.UInt, self._interaction_types.array, tensor_gind),
            shape=(num_valid_paths, depth_dim_size)
        )
        self._vertices_x = mi.TensorXf(
            dr.gather(mi.Float, self._vertices_x.array, tensor_gind),
            shape=(num_valid_paths, depth_dim_size)
        )
        self._vertices_y = mi.TensorXf(
            dr.gather(mi.Float, self._vertices_y.array, tensor_gind),
            shape=(num_valid_paths, depth_dim_size)
        )
        self._vertices_z = mi.TensorXf(
            dr.gather(mi.Float, self._vertices_z.array, tensor_gind),
            shape=(num_valid_paths, depth_dim_size)
        )
        self._shapes = mi.TensorXu(
            dr.gather(mi.UInt, self._shapes.array, tensor_gind),
            shape=(num_valid_paths, depth_dim_size)
        )
        self._primitives = mi.TensorXu(
            dr.gather(mi.UInt, self._primitives.array, tensor_gind),
            shape=(num_valid_paths, depth_dim_size)
        )
        self._buffer_size = num_valid_paths

    def get_interaction_type(self,
                             depth : mi.UInt,
                             active : mi.Bool) -> mi.UInt:
        r"""
        Gathers the interactions types for the specified ``depth``

        :param depth: Depths
        :param active: Flags specifying active components

        :return: Interactions types
        """

        return self._gather_depth(mi.UInt, self._interaction_types, depth,
                                  active)

    def get_vertex(self, depth : mi.UInt, active : mi.Bool) -> mi.Point3f:
        r"""
        Gathers the coordinates of the vertices of the paths for the specified
        ``depth``

        :param depth: Depths
        :param active: Flags specifying active components

        :return: Coordinates of the paths vertices
        """

        vx = self._gather_depth(mi.Float, self._vertices_x, depth, active)
        vy = self._gather_depth(mi.Float, self._vertices_y, depth, active)
        vz = self._gather_depth(mi.Float, self._vertices_z, depth, active)

        return mi.Point3f(vx, vy, vz)

    def get_shape(self, depth : mi.UInt, active : mi.Bool) -> mi.UInt:
        r"""
        Gathers the indices to the intersected shapes for the specified
        ``depth``

        :param depth: Depths
        :param active: Flags specifying active components

        :return: Indices of the intersected shapes
        """

        return self._gather_depth(mi.UInt, self._shapes, depth, active)

    def get_primitive(self, depth : mi.UInt, active : mi.Bool) -> mi.UInt:
        r"""
        Gathers the indices to the intersected primitives for the specified
        ``depth``

        :param depth: Depths
        :param active: Flags specifying active components

        :return: Indices of the intersected primitives
        """

        return self._gather_depth(mi.UInt, self._primitives, depth, active)

    def get_normal(self, depth : mi.UInt, active : mi.Bool) -> mi.Vector3f:
        r"""
        Gathers the normals at the vertices of the paths for the specified
        ``depth``

        :param depth: Depths
        :param active: Flags specifying active components

        :return: Normals at the paths vertices
        """

        # Gather shape pointer for the input depth and cast it to a mesh
        # pointer
        shape_int = self.get_shape(depth, active)
        mesh_ptr = dr.reinterpret_array(mi.MeshPtr, shape_int)

        # Primitive indiex
        prim_ind = self.get_primitive(depth, active)

        # Normal
        normal = mesh_ptr.face_normal(prim_ind, active)
        return normal

    def set_angles_tx(self, k_tx : mi.Vector3f, active : mi.Bool) -> None:
        r"""
        Sets the angles of departure :attr:`~sionna.rt.theta_t`,
        :attr:`~sionna.rt.phi_t` from the directions of departure ``k_tx``

        :param k_tx: Directions of departure
        :param active: Flags specifying active components
        """

        theta_t, phi_t = theta_phi_from_unit_vec(k_tx)
        indices = dr.arange(mi.UInt, self.buffer_size)
        dr.scatter(self._theta_t, theta_t, indices, active)
        dr.scatter(self._phi_t, phi_t, indices, active)

    def set_angles_rx(self, k_rx : mi.Vector3f, active : mi.Bool) -> None:
        r"""
        Sets the angles of arrival :attr:`~sionna.rt.theta_r`,
        :attr:`~sionna.rt.phi_r` from the directions of arrival ``k_rx``

        :param k_rx: Directions of arrival
        :param active: Flags specifying active components
        """

        # Update direction of arrival
        theta_r, phi_r = theta_phi_from_unit_vec(k_rx)
        indices = dr.arange(mi.UInt, self.buffer_size)
        dr.scatter(self._theta_r, theta_r, indices, active)
        dr.scatter(self._phi_r, phi_r, indices, active)

    def set_interaction_type(self,
                             depth : mi.UInt,
                             value : mi.UInt,
                             active : mi.Bool) -> None:
        # pylint: disable=line-too-long
        r"""
        Sets the interactions types for the specified ``depth``

        :param depth: Depths
        :param value: Interactions types represented using :data:`~sionna.rt.constants.InteractionType`
        :param active: Flags specifying active components
        """

        self._scatter_depth(self._interaction_types, depth, value, active)

    def set_vertex(self,
                   depth : mi.UInt,
                   value : mi.Point3f,
                   active : mi.Bool) -> None:
        r"""
        Sets the coordinates of the vertices of the paths for the specified
        ``depth``

        :param depth: Depths
        :param value: Vertices
        :param active: Flags specifying active components
        """

        self._scatter_depth(self._vertices_x, depth, value.x, active)
        self._scatter_depth(self._vertices_y, depth, value.y, active)
        self._scatter_depth(self._vertices_z, depth, value.z, active)

    def set_shape(self,
                  depth : mi.UInt,
                  value : mi.UInt,
                  active : mi.Bool) -> None:
        r"""
        Sets the indices of the intersected shapes for the specified ``depth``

        Invalid intersections should be represented by
        :data:`~sionna.rt.constants.INVALID_SHAPE`

        :param depth: Depths
        :param value: Indices of the intersected shapes
        :param active: Flags specifying active components
        """

        self._scatter_depth(self._shapes, depth, value, active)

    def set_primitive(self,
                      depth : mi.UInt,
                      value : mi.UInt,
                      active : mi.Bool) -> None:
        r"""
        Sets the indices to the intersected primitives for the specified
        ``depth``

        :param depth: Depths
        :param value: Indices of intersected primitives
        :param active: Flags specifying active components
        """

        self._scatter_depth(self._primitives, depth, value, active)

    def detach_geometry(self) -> None:
        r"""
        Detaches the arrays storing the paths geometries (i.e., vertices and
        angles) from the automatic differentiation compuational graph
        """

        self._vertices_x = dr.detach(self._vertices_x)
        self._vertices_y = dr.detach(self._vertices_y)
        self._vertices_z = dr.detach(self._vertices_z)
        self._theta_t = dr.detach(self._theta_t)
        self._phi_t = dr.detach(self._phi_t)
        self._theta_r = dr.detach(self._theta_r)
        self._phi_r = dr.detach(self._phi_r)

    ###############################################
    # Internal methods
    ###############################################

    def _gather_depth(self,
                      dtype : Type[Union[mi.Float, mi.UInt, mi.Bool]],
                      tensor : mi.TensorXf | mi.TensorXu | mi.TensorXb,
                      depth : mi.UInt,
                      active : mi.Bool) -> mi.Float | mi.UInt | mi.Bool:
        r"""
        Gathers data from ``tensor`` for the specified ``depth``

        ``tensor`` is assumed to have shape `[buffer_size, depth_dim_size]`.

        :param dtype: Desired output Mitsuba type
        :param tensor: Tensor to gather from
        :param depth: Depths
        :param active: Flags specifying active components

        :return: Gathered values
        """

        ind = dr.arange(mi.UInt, self.buffer_size)*self.depth_dim_size\
                    + depth - 1
        t = dr.gather(dtype, tensor.array, ind, active)
        return t

    def _scatter_depth(self,
                       tensor : mi.TensorXf | mi.TensorXu | mi.TensorXb,
                       depth : mi.UInt,
                       value : mi.Float | mi.UInt | mi.Bool,
                       active : mi.Bool) -> None:
        r"""
        Sets the items of ``tensor`` to ``value`` for the specified ``depth``

        ``tensor`` is assumed to have shape `[buffer_size, depth_dim_size]`.

        :param tensor: Tensor to update
        :param depth: Depths
        :param value: Value to insert in ``tensor``
        :param active: Flags specifying active components
        """

        ind = dr.arange(mi.UInt, self.buffer_size)*self.depth_dim_size\
                    + depth - 1
        dr.scatter(tensor.array, value, ind, active)
